/* Define if you have a working <inttypes.h> header file.  */
#undef HAVE_INTTYPES_H

/* Whether malloc must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_MALLOC

/* Whether realloc must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_REALLOC

/* Whether calloc must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_CALLOC

/* Whether free must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_FREE

/* Whether index must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_INDEX

/* Whether rindex must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_RINDEX

/* Whether getenv must be declared even if <stdlib.h> is included.  */
#undef NEED_DECLARATION_GETENV
@TOP@
